(**
   This module provides basic data structures for terms and 
   related operations.
   @author Naoki Nishida
   @since 2010/12/17
   @since 2016/08/25
   @since 2016/11/25
   @since 2017/08/02
*)
  
open Util;;
open Variable;;

(**
   This is an exception to indicate that the input is unsupported.
*)
exception Unsupported;;

(**
   This is an exception to indicate that the system is not a normal CTRS.
*)
exception Not_normal_CTRS;;

(**
   This is an exception to indicate that the system is not a oriented 3CTRS.
*)
exception Not_oriented_3CTRS;;

(**
   This is an exception to indicate that the system is not a 3DCTRS.
*)
exception Not_3DCTRS;;

(**
   This is an exception to indicate the existence of contradiction.
*)
exception Contradiction;;

(**
   This is a data type for "CONDITIONTYPE".
*)  
type conditiontype = 
    Oriented 
  | Join 
  | Semiequational;;

(**
   This is a data type for function symbols.
*)
type fid = string * int option;;
    
(**
   This converts [fid] to [string].
*)
let string_of_fid f = string_of_vid f;;

(**
   This converts [fid list] to [string].
*)
let rec string_of_fids =
  function
      [] -> ""
    | [f] -> (string_of_fid f)
    | f::fs -> (string_of_fid f)^" "^(string_of_fids fs);;
	  
(**
   This converts [string] to [fid].
*)
let fid_of_string f = vid_of_string f;;
    
(**
   This is a data structure for symbols.
*)
type symbol = Var of vid | FSym of fid;;
    
(**
   This converts [symbol] to [string].
*)
let string_of_symbol =
  function
      Var x -> string_of_vid x
    | FSym f -> string_of_fid f;;
	  
(**
   This is a data structure for terms.
*)
type term = 
    Node of symbol * term list;;
	  
(**
   This returns the depth of a term.
*)
let rec depth_of_term =
  function
      Node(Var _,_) -> 0
    | Node(f,[]) -> 1
    | Node(f,ts) ->
	1 + depth_of_terms ts
and depth_of_terms =
  function
      [] -> 0
    | t::ts ->
	let n1 = depth_of_term t in
	let n2 = depth_of_terms ts in
	if n1 > n2 then n1 else n2;;
	  
(**
   This decide whether a given term is a variable or not.
*)
let is_var (Node(f,_)) =
  begin
    match f with
      FSym _ -> false
    | Var _ -> true
  end;;
    
(**
   This takes a term and returns variables in the term.
*)
let rec vars_in_term0 ys (Node(f,ts)) =
  let xs = vars_in_terms0 ys ts in
  begin
    match f with
      Var x -> if List.mem x xs then xs else x::xs
    | _ -> xs
  end
and vars_in_terms0 ys =
  function
      [] -> ys
    | t::ts -> 
	let zs = vars_in_term0 ys t in
	vars_in_terms0 zs ts;;
	  
(**
   This takes a term and returns variables in the term.
*)
let vars_in_term t = vars_in_term0 [] t;;
    
(**
   This takes a term list and returns variables in the terms.
*)
let vars_in_terms ts = vars_in_terms0 [] ts;;
    
(**
   This takes a term and returns its root symbol.
*)
let root (Node(f,ts)) = f;;
    
(**
   This converts [term] to [string].
*)
let rec string_of_term (Node(f,ts)) =
  begin
    match ts with
      [] -> string_of_symbol f
    | _ -> (string_of_symbol f)^"("^(string_of_terms ts)^")"
  end
(**
   This converts [term list] to [string].
*)
and string_of_terms =
  function
      [] -> ""
    | [t] -> string_of_term t
    | t::ts -> (string_of_term t)^","^(string_of_terms ts);;
	  
(**
   This is a data structure for conditions.
*)
type condition = 
    Arrow of term * term 
  | Darrow of term * term
  | Equal of term * term;;
    
(**
   This converts [condition] to [string].
*)
let string_of_condition ct c =
  begin
    match ct with
      None ->
	begin
	  match c with
	    Arrow(s,t) -> (string_of_term s)^" -> "^(string_of_term t)
	  | Darrow(s,t) -> (string_of_term s)^" -><- "^(string_of_term t)
	  | Equal(s,t) -> (string_of_term s)^" == "^(string_of_term t)
	end
    | Some _ ->
	begin
	  match c with
	    Arrow(s,t) -> (string_of_term s)^" == "^(string_of_term t)
	  | Darrow(s,t) -> (string_of_term s)^" == "^(string_of_term t)
	  | Equal(s,t) -> (string_of_term s)^" == "^(string_of_term t)
	end
  end
	  
(**
   This takes a condtion and returns variables in the condtion.
*)
let vars_in_condition0 ys =
  function
      Arrow (s,t) -> vars_in_terms0 ys [s;t]
    | Darrow (s,t) -> vars_in_terms0 ys [s;t]
    | Equal (s,t) -> vars_in_terms0 ys [s;t];;
	  
(**
   This takes a condition list and returns variables in the condtions.
*)
let rec vars_in_conditions0 ys =
  function
      [] -> ys
    | c::cs ->
	let zs = vars_in_condition0 ys c in
	vars_in_conditions0 zs cs;;

(**
   This takes the lhs of a given condition.
*)
let lhs_of_condition =
  function
      Arrow (s,t) -> s
    | Darrow (s,t) -> s
    | Equal (s,t) -> s;;
	  
(**
   This takes the rhs of a given condition.
*)
let rhs_of_condition =
  function
      Arrow (s,t) -> t
    | Darrow (s,t) -> t
    | Equal (s,t) -> t;;
	  
(**
   This takes a condtion and returns variables in the condtion.
*)
let vars_in_condition c = vars_in_condition0 [] c;;
    
(**
   This takes a condition list and returns variables in the condtions.
*)
let vars_in_conditions c = vars_in_conditions0 [] c;;
    
(**
   This is a data structure for conditional parts.
*)
type conditions = condition list;;
    
(**
   This converts [conditions] to [string].
*)
let rec string_of_conditions ct =
  function
      [] -> ""
    | [c] -> string_of_condition ct c
    | c::cs -> (string_of_condition ct c)^", "^(string_of_conditions ct cs);;
	  
(**
   This is a data structure for conditional rewrite rules.
*)
type rule = 
    Rew of term * term * conditions
  | RelRew of term * term * conditions;;
	
(**
   This converts [rule] to [string].
*)
let rec string_of_rule ct =
  function
      Rew(l,r,c) ->
	(string_of_term l)^" -> "^(string_of_term r)
	^
	begin
	  match c with
	    [] -> ""
	  | _ -> " | "^(string_of_conditions ct c)
	end
    | RelRew(l,r,c) ->
	(string_of_term l)^" ->= "^(string_of_term r)
	^
	begin
	  match c with
	    [] -> ""
	  | _ -> " | "^(string_of_conditions ct c)
	end;;
	  
(**
   This takes a rule and returns its lhs.
*)
let lhs =
  function
      Rew(l,_,_) -> l
    | RelRew(l,_,_) -> l;;
	  
(**
   This takes a rule and returns its rhs.
*)
let rhs =
  function
      Rew(_,r,_) -> r
    | RelRew(_,r,_) -> r;;
	  
(**
   This returns the depth of the lhs of a rule.
*)
let depth_of_lhs_in_rule rl = depth_of_term (lhs rl);;
    
(**
   This returns the maximum depth of lhss in rules.
*)
let rec depth_of_lhs_in_rules =
  function
      [] -> 0
    | rl::rs -> 
	let n1 = depth_of_lhs_in_rule rl in
	let n2 = depth_of_lhs_in_rules rs in
	if n1 > n2 then n1 else n2;;
	  
(**
   This takes a rule and returns variables in the rule.
*)
let vars_in_rule =
  function
      Rew(l,r,c) -> union (vars_in_terms [l;r]) (vars_in_conditions c)
    | RelRew(l,r,c) -> union (vars_in_terms [l;r]) (vars_in_conditions c);;

(**
   This takes a rule list and returns variables in the rules.
*)
let rec vars_in_rules =
  function
      [] -> []
    | (r::rs) ->
	union (vars_in_rule r) (vars_in_rules rs);;
	  
(**
   This is a data structure for sets of rewrite rules.
*)
type rules = rule list;;
	  
(*
   This converts [rules] to [string] with [n] indents.
*)
let rec string_of_rules n ct =
  function
      [] -> ""
    | r::rs -> 
	(string_of_indent n)^(string_of_rule ct r)
	^"\n"
	^(string_of_rules n ct rs);;
	  
(**
   This is a data structure for unconditional equations.
*)
type equation = term * term;;
	  
(**
   This converts [equation] to [string].
*)
let string_of_equation (s,t) =
  (string_of_term s)^" == "^(string_of_term t);;
    
(**
   This converts [equation list] to [string] with [n] indents.
*)
let rec string_of_equations n =
  function
      [] -> ""
    | e::es ->
	(string_of_indent n)^(string_of_equation e)^"\n"
	^(string_of_equations n es);;

(**
   This is a data structure for theories.
*)
type theory = Builtin of string * fid list | Equations of equation list;;
    
(**
   This converts [theory] to [string] with [n] indents.
*)
let string_of_theory n =
  function
      Builtin (f,fs) -> 
	(string_of_indent n)^"("^f^" "^(string_of_fids fs)^")"
    | Equations es -> 
	(string_of_indent n)^"(EQUATIONS\n"^(string_of_equations (n+1) es)^")";;
	  
(**
   This converts [theory list] to [string] with [n] indents.
*)
let rec string_of_theories n =
  function
      [] -> ""
    | th::ths -> (string_of_theory n th)^"\n"^(string_of_theories n ths);;
	  
(**
   This is a data structure for context-sensitive replacements.
*)
type csreplace = fid * int list;;
	  
(**
   This converts [int list] to [string].
*)
let rec string_of_intlist =
  function
      [] -> ""
    | [n] -> string_of_int n
    | n::ns -> (string_of_int n)^" "^(string_of_intlist ns);;
	  
(**
   This converts [csreplace] to [string].
*)
let string_of_csreplace (f,ns) =
  "("^(string_of_fid f)^" "^(string_of_intlist ns)^")";;
    
(**
   This is a data structure for context-sensitive replacement mappings.
*)
type csmapping = csreplace list;;
    
(**
   This converts [csmapping] to [string] with [n] indents.
*)
let rec string_of_csmapping n =
  function
      [] -> ""
    | cs::css -> 
	(string_of_indent n)^(string_of_csreplace cs)^"\n"
	^(string_of_csmapping n css);;
	  
(**
   This is a data structure for reduction strategies.
*)
type strategy = Innermost | Outermost | ContextSensitive of csmapping;;
  
(**
   This converts [strategy] to [string] with [n] indents.
*)
let string_of_strategy n =
  function
      Innermost -> (string_of_indent n)^"INNERMOST"
    | Outermost -> (string_of_indent n)^"OUTERMOST"
    | ContextSensitive css -> 
	(string_of_indent n)^"CONTEXTSENSITIVE\n"
	^(string_of_csmapping (n+1) css);;
	  
(**
   This is a data structure for other declarations.
*)
type otherdecl = string * string list;;
	  
(**
   This converts [string list] to [string].
*)
let rec string_of_strings =
  function
      [] -> ""
    | [s] -> s
    | s::ss -> s^" "^(string_of_strings ss);;
	  
(**
   This converts [otherdecl] to [string].
*)
let string_of_otherdecl (s,ss) = s^" "^(string_of_strings ss);;
    
(**
   This converts [otherdecl list] to [string].
*)
let rec string_of_otherdecls n =
  function
      [] -> ""
    | d::ds -> 
	(string_of_indent n)^"("^(string_of_otherdecl d)^")\n"^(string_of_otherdecls n ds);;
	  
(**
   This is a data structure for CTRSs obtained from [.trs] files.
*)
type ctrs = 
    { var : vid list;
      theories : theory list;
      rules : rules;
      strategy : strategy option;
      others : otherdecl list;
      ctype : conditiontype option;
    };;

(**
   This computes the maximum number of conditions in CTRS.
*)
let rec max_of_conditions =
  function
      [] -> 0
    | rl::rs ->
	let c = 
	  begin
	    match rl with
	      Rew(_,_,c) -> c
	    | RelRew(_,_,c) -> c
	  end 
	in
	let n = max_of_conditions rs in
	let m = List.length c in
	if n > m then n else m;;
	  
(**
   This converts [ctrs] to [string].
*)
let string_of_ctrs n spec =
  begin
    match spec.ctype with
      None -> ""
    | Some c ->
	(string_of_indent n)^"(CONDITIONTYPE "
	^
	begin
	  match c with
	    Semiequational -> "SEMI-EQUATIONAL"
	  | Join -> "JOIN"
	  | Oriented -> "ORIENTED"
	end
	^
	")\n"
  end
  ^
  begin
    match spec.var with
      [] -> ""
    | _ -> (string_of_indent n)^"(VAR "^(string_of_vids spec.var)^")\n"
  end
  ^
  begin
    match spec.theories with
      [] -> ""
    | _ -> 
	(string_of_indent n)
	^"(THEORY\n"^(string_of_theories (n+1) spec.theories)
	^(string_of_indent n)^")\n"
  end
  ^
  begin
    match spec.rules with
      [] -> ""
    | _ -> 
	(string_of_indent n)
	^"(RULES\n"^(string_of_rules (n+1) spec.ctype spec.rules)
	^(string_of_indent n)^")\n"
  end
  ^
  begin
    match spec.strategy with
      None -> ""
    | Some st -> "(STRATEGY "^(string_of_strategy 1 st)^")\n"
  end
  ^
  (string_of_otherdecls n spec.others);;
    
(**
   This converts [ctrs] to [string] in AProVE format.
*)
let string_of_ctrs_for_aprove n spec =
  begin
    match spec.var with
      [] -> ""
    | _ -> (string_of_indent n)^"["^(string_of_vids_with_comma spec.var)^"]\n"
  end
  ^
  begin
    match spec.theories with
      [] -> ""
    | _ -> raise (Failure "unexpected")
  end
  ^
  begin
    match spec.rules with
      [] -> ""
    | _ -> 
	(string_of_indent n)
	^(string_of_rules (n+1) spec.ctype spec.rules)
  end
  ^
  begin
    match spec.strategy with
      None -> ""
    | Some st -> raise (Failure "unexpected")
  end;;
    
(**
   This constructs a [ctrs] data from a [rules] data.
*)
let mkFromRules rs =
  let xs = vars_in_rules rs in
  { var = xs; 
   theories = []; 
   rules = rs; 
   strategy = None;
   others = []; 
   ctype = None;
   };;
    
(*
   This takes a [rules] and returns the list of the defined symbols.
*)
let rec get_dsym0 ds =
  function
      [] -> ds
    | rl::rs ->
	let l1 =
	  begin
	    match rl with
	      Rew(l,r,c) -> l
	    | RelRew(l,r,c) -> l
	  end
	in
	begin
	  match root l1 with
	    Var _ -> raise (Failure "Error: there is a variable lhs.\n")
	  | FSym f -> 
	      if List.mem f ds then get_dsym0 ds rs else get_dsym0 (f::ds) rs
	end;;
	  
(**
   This takes a [rules] and returns the list of the defined symbols.
*)
let get_dsym rs = get_dsym0 [] rs;;
    
(**
   This takes a rule list and an [fid], and returns a list of [fid]-rules.
*)
let rec get_rules f =
  function
      [] -> []
    | (rl::rs) ->
	let g = root (lhs rl) in
	begin
	  match g with
	    Var _ -> raise (Failure "Error: there is a variable lhs.\n")
	  | FSym g1 ->
	      if f = g1 then rl::(get_rules f rs) else get_rules f rs
	end;;	      
	  
(**
   This takes a CTRS and an [fid], and returns a list of [fid]-rules 
   in the CTRS.
*)
let rules spec f = get_rules f spec.rules;;
    
(**
   This computes symbols in the term that belong to the input symbol list.
*)
let rec fids_in_term fs (Node(f,ts)) =
  begin
    match f with
      Var _ -> []
    | FSym g ->
	if List.mem g fs 
	then 
	  let gs = fids_in_terms fs ts in
	  if List.mem g gs then gs else g::gs
	else fids_in_terms fs ts
  end
and fids_in_terms fs =
  function
      [] -> []
    | t::ts -> union (fids_in_term fs t) (fids_in_terms fs ts);;
	  
(**
   This computes symbols in the conditions that belong to the input symbol list.
*)
let rec fids_in_conditions fs =
  function
      [] -> []
    | c::cs ->
	let ts = 
	  begin
	    match c with
	      Arrow(s,t) -> [s;t]
	    | Darrow(s,t) -> [s;t]
            | Equal(s,t) -> [s;t]
	  end
	in
	union (fids_in_terms fs ts) (fids_in_conditions fs cs);;

(**
   This computes all symbols in a term.
*)
let rec allfids_in_term =
  function
      Node(Var _,_) -> []
    | Node(FSym f,ts) -> 
	let fs = allfids_in_terms ts in
	if List.mem f fs then fs else f::fs
and allfids_in_terms =
  function
      [] -> []
    | t::ts -> union (allfids_in_term t) (allfids_in_terms ts);;
	  
(**
   This computes all symbols in a condition.
*)
let allfids_in_condition =
  function
      Arrow(s,t) -> allfids_in_terms [s;t]
    | Darrow(s,t) -> allfids_in_terms [s;t]
    | Equal(s,t) -> allfids_in_terms [s;t];;
	  
(**
   This computes all symbols in conditions.
*)
let rec allfids_in_conditions =
  function 
      [] -> []
    | c::cs -> union (allfids_in_condition c) (allfids_in_conditions cs);;

(**
   This computes symbols in a rule.
*)
let allfids_in_rule =
  function
      Rew(l,r,c) -> union (allfids_in_terms [l;r]) (allfids_in_conditions c)
    | RelRew(l,r,c) ->
	union (allfids_in_terms [l;r]) (allfids_in_conditions c);;

(**
   This computes symbols in rules.
*)
let rec allfids_in_rules =
  function
      [] -> []
    | rl::rs -> union (allfids_in_rule rl) (allfids_in_rules rs);;

(**
   This computes the arity table for defined symbols in [spec].
*)
let dsym_arity_table rs = 
  let rec compute_arity table =
    function
	[] -> table
      |	rl::rs -> 
	  let lhs = 
	    begin
	      match rl with
		Rew(l,_,_) -> l
	      |	RelRew(l,_,_) -> l
	    end
	  in
	  begin
	    match lhs with
	      Node(Var _,_) -> compute_arity table rs
	    | Node(FSym f,ts) ->
		try
		  let k = List.assoc f table in
		  let m = List.length ts in
		  if m = k then compute_arity table rs
		  else
		    raise 
		      (Failure 
			 ("Error: the arity of "^(string_of_fid f)^" is not fixed.\n"))
		with
		  Not_found -> compute_arity ((f,List.length ts)::table) rs
	  end   
  in
  compute_arity [] rs;;


(**
   This generates a funciton that returns arity of the input defined symbol.
*)
let mk_dsymarity_function spec =
  (fun g -> List.assoc g (dsym_arity_table spec.rules));;
    
(**
   This is a data type for substitutions.
*)
type subst = (vid * term) list;;

(**
    This is a data type for variable renamings.
*)
type renaming = subst;;
    
(**
   This is an identity substitution.
*)
let identity_subst = ([] : subst);;
    
(**
   This converts [subst] to [string].
*)
let string_of_subst theta =
  let rec string_of_subst0 =
    function 
	[] -> " }"
      |	[(x,t)] ->
	  (string_of_vid x)^" -> "^(string_of_term t)^(string_of_subst0 [])
      |	(x,t)::sigma -> 
	  (string_of_vid x)^" -> "^(string_of_term t)^", "
	  ^(string_of_subst0 sigma)
  in
  "{ "^(string_of_subst0 theta);;
    
(**
   This converts [subst list] to [string].
*)
let rec string_of_substs =
  function
      [] -> ""
    | [theta] -> string_of_subst theta
    | theta::thetas -> (string_of_subst theta)^", "^(string_of_substs thetas);;
    
(**
   This computes the domain of substitution [theta].
*)
let rec domain =
  function
      [] -> []
    | (x,t)::theta -> x::(domain theta);;
	  
(**
   This decides whether a substitution is a variable renaming or not.
*)
let is_renaming xs0 theta =
  let xs1 = (Util.setminus xs0 (domain theta)) in
  let rec is_renaming0 xs =
    function 
	[] -> true
      |	(x,Node(Var y,_))::sigma ->
	  if List.mem y xs
	  then false
	  else is_renaming0 (y::xs) sigma
      |	_ -> false
  in
  is_renaming0 xs1 theta;;
    
(**
   This applys a substitution [theta] to a term [term].
*)
       let rec apply_subst_to_term theta t =
  begin
    match t with
      Node(Var x,_) -> 
	begin
	  try
	    List.assoc x theta
	  with
	    Not_found -> t
	end
    | Node(f,ts) ->
	Node(f,List.map (apply_subst_to_term theta) ts)
  end;;

(**
   This applys a substitution [theta] to a condition.
*)
let apply_subst_to_condition theta =
  function
      Arrow(t1,t2) -> 
	Arrow(apply_subst_to_term theta t1, apply_subst_to_term theta t2)
    | Darrow(t1,t2) ->
	Darrow(apply_subst_to_term theta t1, apply_subst_to_term theta t2)
    | Equal(t1,t2) ->
	Equal(apply_subst_to_term theta t1, apply_subst_to_term theta t2);;
	  
(**
   This restricts a substitution [theta] to a variable set [xs].
*)
let rec restrict_subst xs theta =
  begin
    match xs with
      [] -> []
    | x::xs1 -> 
	begin
	  try
	    (x,List.assoc x theta)::(restrict_subst xs1 theta)
	  with
	    Not_found -> restrict_subst xs1 theta
	end
  end;;
    
(**
   This combines substituitons.
*)
let rec combine_subst theta =
  function
      [] -> theta
    | ((x,t)::sigma) ->
	begin
	  try
	    let u = List.assoc x theta in
	    if t = u 
	    then combine_subst theta sigma
	    else raise Not_found
	  with
	    Not_found -> combine_subst ((x,t)::theta) sigma
	end;;

(**
   This composes substitutions.
*)
let compose_subst theta sigma =
   let theta1 = List.map (fun (x,t) -> (x, apply_subst_to_term sigma t)) theta
   in
   let xs = setminus (domain sigma) (domain theta) in
   let sigma1 = restrict_subst xs sigma in
   try
     combine_subst theta1 sigma1
   with
     Not_found -> raise (Failure "Error in compose_subst: unexpected.\n");;
      
(**
   This applys a substitution [theta] to conditions.
*)
let rec apply_subst_to_conditions theta =
  function
      [] -> []
    | c::cs -> 
	(apply_subst_to_condition theta c)
	::(apply_subst_to_conditions theta cs);;

(**
    This applys a substitution [theta] to a rule [rule].
*)
let apply_subst_to_rule theta =
  function
      Rew(l,r,c) -> Rew(apply_subst_to_term theta l,
			apply_subst_to_term theta r,
			apply_subst_to_conditions theta c)
    | RelRew(l,r,c) -> RelRew(apply_subst_to_term theta l,
			      apply_subst_to_term theta r,
			      apply_subst_to_conditions theta c);;

(**
   This is a data type for positions in terms.
*)
type position = int list;;

(**
   This is a root position of terms.
*)
let root_position = ([] : position);;

(**
   This converts [position] to [string].
*)
let rec string_of_position =
  function
      [] -> "e"
    | [n] -> string_of_int n
    | n::ns -> (string_of_int n)^"."^(string_of_position ns);;

(**
   This is a data type for position sets.
*)
type positions = position list;;

(**
   This converts [positions] to [string].
*)
let rec string_of_positions =
  function
      [] -> ""
    | [p] -> string_of_position p
    | p::ps -> (string_of_position p)^", "^(string_of_positions ps);;

(**
   This returns variable positions in a term.
*)
let rec vpos_in_term =
  function
      Node(Var x,_) -> [[]]
    | Node(f,ts) -> vpos_in_terms 1 ts
and vpos_in_terms n =
  function
      [] -> []
    | (t::ts) ->
	(List.map (fun p -> n::p) (vpos_in_term t))
	@ (vpos_in_terms (n+1) ts);;

(**
   This checks whether [p] is a prefix of [ps].
*)
let rec is_prefix p1 p2 =
  begin
    match (p1,p2) with
      ([],_) -> true
    | (p1h::p1tl,p2h::p2tl) -> p1h = p2h && is_prefix p1tl p2tl
    | _ -> false
  end;;
    
(**
   This removes all postifx of [p] from [ps].
*)
let rec remove_postfix p =
  function
      [] -> []
    | p1::ps1 -> 
	if is_prefix p p1
	then remove_postfix p ps1 
	else p1::remove_postfix p ps1;;
	  
(**
   This is an exception returned by matching functions when the input 
   is not matchable.
*)
exception Not_matching;;
	  
(**
   This is a matching of terms.
*)
let rec match_term t u =
  begin
    match t with
      Node(Var x,_) -> [(x,u)]
    | Node(f,ts) ->
	begin
	  match u with
	    Node(Var _,_) -> raise Not_matching
	  | Node(g,us) -> 
	      if f = g 
	      then match_terms ts us
	      else raise Not_matching
	end
  end
and match_terms ts0 us0 = 
  begin
    match (ts0,us0) with
      ([],[]) -> []
    | (t::ts,u::us) ->
	begin
	  try
	    combine_subst (match_term t u) (match_terms ts us)
	  with
	    Not_found -> raise Not_matching
	end
    | _ -> raise (Failure "Error in match_terms: unexpected.\n")
  end;;    
    
(**
   This decides whether two conditions are matchable or not.
*)
let match_condition c1 c2 =
  begin
    match (c1,c2) with
      (Arrow(t1,u1),Arrow(t2,u2)) -> match_terms [t1;u1] [t2;u2]
    | (Darrow(t1,u1),Darrow(t2,u2)) -> match_terms [t1;u1] [t2;u2]
    | _ -> raise Not_matching
  end;;
    
(**
   This decides whether two condition lists are matchable or not.
*)
let rec match_conditions cs1 cs2 =
  begin
    match (cs1,cs2) with
      ([],[]) -> []
    | (c3::cs3,c4::cs4) ->
	begin
	  try
	    combine_subst (match_condition c3 c4) (match_conditions cs3 cs4)
	  with
	    Not_found -> raise Not_matching
	end
    | _ -> raise Not_matching
  end;;
    
(**
   This decides whether two rewrite rules are matchable or not.
*)
let match_rule rl1 rl2 =
  begin
    match (rl1,rl2) with
      (Rew(l1,r1,c1),Rew(l2,r2,c2)) ->
	begin
	  try
	    combine_subst (match_terms [l1;r1] [l2;r2]) (match_conditions c1 c2)
	  with
	    Not_found -> raise Not_matching
	end
    | (RelRew(l1,r1,c1),RelRew(l2,r2,c2)) ->
	begin
	  try
	    combine_subst (match_terms [l1;r1] [l2;r2]) (match_conditions c1 c2)
	  with
	    Not_found -> raise Not_matching
	end
    | _ -> raise Not_matching
  end;;

(**
   This is a auxiliary function for [unify].
*)
let rec usub alist (Node(f1,ts1) as t1) (Node(f2,ts2) as t2) =
  begin
    match t1 with
      Node(Var x,_) -> 
	if f1 = f2
	then
	  alist
	else
	  if List.mem x (vars_in_term t2)
	  then 
	    raise Not_found
	  else
	    (x,t2)::(List.map 
		       (fun (y,u) -> (y, apply_subst_to_term [(x,t2)] u))
		       alist)
    | _ ->
	begin
	  match t2 with
	    Node(Var x,_) -> 
	      if List.mem x (vars_in_term t1) 
	      then
		raise Not_found
	      else 
		(x,t1)::(List.map
			   (fun (y,u) -> (y, apply_subst_to_term [(x,t1)] u))
			   alist)
	  | _ -> 
	      if f1 = f2 
	      then 
		usublist alist ts1 ts2
	      else
		raise Not_found
	end
  end
and usublist alist ts1 ts2 =
  begin
    match (ts1,ts2) with
      ([],[]) -> alist
    | (u1::us1,u2::us2) -> 
	let alist1 = usub alist u1 u2 in
	usublist alist1 
	  (List.map (apply_subst_to_term alist1) us1)
	  (List.map (apply_subst_to_term alist1) us2)
    | _ -> raise Not_found
  end;;
	
(**
   This unifies two terms.
*)
let unify t1 t2 = usub [] t1 t2;;
    
(**
   This decides whether two terms are unifiable or not.
*)
let is_unifiable t1 t2 =
  begin
    try
      let _ = unify t1 t2 in
      true
    with
      Not_found -> false
  end;;

(**
   This decides whether two rewrite rules are variants each other or not.
*)
let is_variant rl1 rl2 = 
  begin
    try
      let theta = match_rule rl1 rl2 in
      let xs = vars_in_rule rl1 in
      is_renaming xs theta
    with
      Not_matching -> false
  end;;

(**
   This decides whether a substitution is more general than another.
*)
let more_general_subst theta sigma =
  let rec more_general_subst0 delta =
    function
	[] -> true
      |	(x,t)::theta1 ->
	  if root t = Var x
	  then more_general_subst0 delta theta1 
	  else 
	    begin
	      try
		let u = List.assoc x sigma in
		let delta1 = match_term t u in
		more_general_subst0 (combine_subst delta delta1) theta1
	      with
		Not_found -> false
	      |	Not_matching -> false
	    end
  in
  more_general_subst0 identity_subst theta;;
    
(**
   This decides whether [t] is a subterm of [u].
*)
let rec is_subterm t u =
  begin
    match u with
      Node(f,[]) -> t = u
    | Node(f,us) -> t = u || List.exists (is_subterm t) us
  end;;
    
(**
   This decides whether [t] is a proper subterm of [u].
*)
let is_proper_subterm t =
  function
      Node(Var _,_) -> false
    | Node(FSym _,ts) ->
	List.exists (is_subterm t) ts;;

(**
   This decides whether a given CTRS is unconditional.
*)
let is_unconditional spec =
  let rec uncond =
    function
	[] -> true
      |	(Rew(_,_,[]))::rs -> uncond rs
      |	(RelRew(_,_,[]))::rs -> uncond rs
      |	_ -> false
  in
  uncond spec.rules;;

(**
   This decides whether a given CTRS has a rule of [RelRew].
*)
let has_relrew spec = 
   let rec relrew =
   function
   [] -> false
   | (Rew _)::rs -> relrew rs
   | (RelRew _)::_ -> true
   in
   relrew spec.rules;;

(**
   This decides whether a given term is linear.
*)
let is_linear_term t =
  let vs = ref [] in
  let rec linear_term =
    function
	Node(Var x,_) -> 
	  if List.mem x !vs 
	  then false
	  else 
	    begin
	      vs := x::(!vs);
	      true
	    end
      | Node(FSym _,ts) -> linear_terms ts
  and linear_terms =
    function
	[] -> true
      | t::ts -> (linear_term t) && (linear_terms ts)
  in
   linear_term t;;
    
(**
   This decides whether a given CTRS is left-linear.
*)
let is_LL spec =
  let rec leftlinear_rules =
    function
	[] -> true
      | (Rew(l,_,_))::rs -> is_linear_term l && leftlinear_rules rs
      | (RelRew(l,_,_))::rs -> is_linear_term l && leftlinear_rules rs
  in
  leftlinear_rules spec.rules;;

(**
    This collects non-linear variables in a term.
*)
let nonlinear_vars_in_term t =
  let vs1 = ref [] in
  let vs2 = ref [] in
  let rec get_vids =
    function
	Node(Var x,_) ->
	  if List.mem x !vs1 
	  then if List.mem x !vs2 then () else vs2 := x::(!vs2)
	  else vs1 := x::(!vs1)
      |	Node(FSym _,ts) ->
	  let _ = List.map get_vids ts in
	  ()
  in
  let _ = get_vids t in
  !vs2;;

(**
    This collects non-linear variables in a list of terms.
*)
let nonlinear_vars_in_terms ts = nonlinear_vars_in_term (Node(FSym ("#dummy",None),ts));;
   
(**
   This decides whether a given term(s) is ground.
*)
let rec is_ground_term =
  function
      Node(Var _,_) -> false
    | Node(FSym _,ts) -> is_ground_terms ts
and is_ground_terms =
  function
      [] -> true
    | t::ts -> is_ground_term t && is_ground_terms ts;;

let rec is_lhs_instance rs t =
    match rs with
      [] -> false
    | (Rew(l,_,_))::rs ->
	begin
	  try
	    let _ = match_term l t in
	    true
	  with
	    Not_matching -> is_lhs_instance rs t
	end
    | (RelRew(l,_,_))::rs ->
	begin
	  try
	    let _ = match_term l t in
	    true
	  with
	    Not_matching -> is_lhs_instance rs t
	end;;

(**
   This decides whether a given term(s) is ground strongly irreducible.
   Note that this is not complete.
*)
let rec is_ground_strongly_irreducible_term spec t =
  begin
    match t with
      Node(Var _,_) -> false
    | Node(FSym _,ts) ->
	not (is_lhs_instance spec.rules t)
	  && is_ground_strongly_irreducible_terms spec ts
  end
and is_ground_strongly_irreducible_terms spec =
  function
      [] -> true
    | t::ts ->
	is_ground_strongly_irreducible_term spec t
	  && is_ground_strongly_irreducible_terms spec ts;;

(**
   This decides whether a given CTRS is Type 1.
*)
let is_type_one spec =
  let rec is_type1_rules =
    function
	[] -> true
      |	(Rew (l,r,c))::rs -> 
	  is_subseteq (union (vars_in_term r) (vars_in_conditions c)) (vars_in_term l)
	    && is_type1_rules rs
      |	(RelRew (l,r,c))::rs -> 
	  is_subseteq (union (vars_in_term r) (vars_in_conditions c)) (vars_in_term l)
	    && is_type1_rules rs
  in
  is_type1_rules spec.rules;;
  
(**
   This decides whether a given CTRS is Type 3.
*)
let is_type_three spec =
  let is_type3_rule rl =
    let (l,r,c) =
      begin
	match rl with
	  Rew(l,r,c) -> (l,r,c)
	| RelRew(l,r,c) -> (l,r,c)
      end
    in
    is_subseteq (vars_in_term r) (union (vars_in_conditions c) (vars_in_term l))
  in
  let rec is_type3_rules =
    function
	[] -> true
      |	rl::rs -> 
	  is_type3_rule rl && is_type3_rules rs
  in
  is_type3_rules spec.rules;;
  
(**
   This decides whether a given CTRS is normal.
*)
let is_normal spec =
  let rec is_normal_cond =
    function
	[] -> true
      |	(Arrow (s,t))::cs ->
	  is_ground_strongly_irreducible_term spec t
	    && is_normal_cond cs
      |	_::_ -> false
  and is_normal_rules =
    function
	[] -> true
      |	(Rew(l,r,c))::rs ->
	  is_normal_cond c && is_normal_rules rs
      |	(RelRew(l,r,c))::rs ->
	  is_normal_cond c && is_normal_rules rs
  in
  is_normal_rules spec.rules;;

(**
   This decides whether a given rule is deterministic.
*)
let rec is_deterministic_rule rl =
  let rec is_D vs cs =
    begin
      match cs with
	[] -> true
      |	(Arrow (s,t))::cs1 ->
	  let xs = vars_in_term s in
	  if is_subseteq xs vs 
	  then
	    is_D (union vs (vars_in_term t)) cs1
	  else
	    false
      |	_ -> raise (Failure "unexpected use of is_deterministic_rule")
    end
  in
  let (l,r,c) =
    begin
      match rl with
	Rew(l,r,c) -> (l,r,c)
      |	RelRew(l,r,c) -> (l,r,c)
    end
  in
  is_D (vars_in_term l) c;;

(**
   This decides whether a given CTRS is deterministic.
*)
let is_deterministic spec =
  List.for_all is_deterministic_rule spec.rules;;

(**
   This decides whether a given CTRS is a 3-DCTRS.
*)
let is_3DCTRS spec = is_deterministic spec && is_type_three spec;;

(**
   This decides whether a given CTRS is non-erasing.
*)
let rec is_NE_rule rl =
  let (l,r,c) = 
    begin
      match rl with
	Rew(l,r,c) -> (l,r,c)
      |	RelRew(l,r,c) -> (l,r,c)
    end
  in
  is_subseteq (vars_in_term l) (vars_in_term r)
and is_NE spec = List.for_all is_NE_rule spec.rules;;

(**
   This decides whether a given CTRS has ground conditions only.
*)
let rec is_groundconditions_rule rl =
  let (l,r,c) = 
    begin
      match rl with
	Rew(l,r,c) -> (l,r,c)
      |	RelRew(l,r,c) -> (l,r,c)
    end
  in
  vars_in_conditions c = []
and is_groundconditions spec = List.for_all is_NE_rule spec.rules;;

(**
   This decides whether a given CTRS is weakly left-linear.
*)
let rec is_WLL_rule rl =
  let (l,r,c) = 
    begin
      match rl with
	Rew(l,r,c) -> (l,r,c)
      |	RelRew(l,r,c) -> (l,r,c)
    end
  in
  begin
    let nlvids = nonlinear_vars_in_terms (l::(List.map rhs_of_condition c)) in
    let rvids = vars_in_terms (r::(List.map lhs_of_condition c)) in
    List.for_all (fun x -> not (List.mem x rvids)) nlvids
  end
and is_WLL spec = List.for_all is_WLL_rule spec.rules;;

(**
   This decides whether a given CTRS is ultra-weakly-left-linear.
*)
let rec is_UWLL_rule rl =
  let (l,r,c) = 
    begin
      match rl with
	Rew(l,r,c) -> (l,r,c)
      |	RelRew(l,r,c) -> (l,r,c)
    end
  in
  begin
    match c with
      [] -> 
	let nlvids = nonlinear_vars_in_term l in
	let rvids = vars_in_term r in
	List.for_all (fun x -> not (List.mem x rvids)) nlvids
    | _ -> 
	let ts = List.rev (List.map rhs_of_condition c) in
	let ts1 = List.tl ts in
	let nlvids = nonlinear_vars_in_terms (l::ts) in
	let rvids = vars_in_term r in
	is_linear_term (Node(FSym ("#dummy",None),l::ts1))
	  && List.for_all (fun x -> not (List.mem x rvids)) nlvids
  end
and is_UWLL spec = List.for_all is_UWLL_rule spec.rules;;

(**
   This decides whether a given CTRS is ultra-weakly-left-linear.
*)
let rec is_URL_rule rl =
  let (l,r,c) = 
    begin
      match rl with
	Rew(l,r,c) -> (l,r,c)
      |	RelRew(l,r,c) -> (l,r,c)
    end
  in
  is_linear_term r && (vars_in_terms (List.map lhs_of_condition c) = [])
and is_URL spec = List.for_all is_URL_rule spec.rules;;

(**
   This decides whether a given CTRS is standard (no strategy, no theory).
*)
let is_standard spec = 
  match (spec.theories, spec.strategy) with
    ([],None) -> true
  | _ -> false;;

(**
   This collects variable occurrences in a given term.
*)
let get_var_occ t =
  let vs = ref [] in
  let rec insert x =
    function
	[] -> [(x,1)]
      |	((y,m)::vs1) ->
	  if x = y
	  then (y,m+1)::vs1
	  else (y,m)::(insert x vs1)
  in
  let rec get_vocc =
    function
	Node(Var x,_) -> vs := insert x (!vs)
      |	Node(FSym _,ts) ->
	  let _ = List.map get_vocc ts in
	  ()
  in
  get_vocc t;
  !vs;;
    
(**
   This decides whether a unconditional rule is non-duplicating.
*)
let is_nonduplicating =
  function
      Rew(l,r,[]) ->
	begin
	  let vocc1 = get_var_occ l in
	  let vocc2 = get_var_occ r in
	  let lte (x,m) =
	    try
	      let n = List.assoc x vocc1 in
	      n >= m
	    with
	      Not_found -> false
	  in
	  List.for_all lte vocc2
	end
    | RelRew(l,r,[]) ->
	begin
	  let vocc1 = get_var_occ l in
	  let vocc2 = get_var_occ r in
	  let lte (x,m) =
	    try
	      let n = List.assoc x vocc1 in
	      n >= m
	    with
	      Not_found -> false
	  in
	  List.for_all lte vocc2
	end
    | _ -> false;;
	  
(**
   This computes the size of a given term.
*)
let rec size_of_term =
  function
      Node(Var _,_) -> 1
    | Node(FSym _,ts) -> 1 + size_of_terms ts
and size_of_terms =
  function
      [] -> 0
    | t::ts -> size_of_term t + size_of_terms ts;;
	  
(**
   This computes the size of a given term using a given weight function.
*)
let rec size_of_term_with_weight w =
  function
      Node(Var _,_) -> 1
    | Node(FSym f,ts) -> (w f) + size_of_terms_with_weight w ts
and size_of_terms_with_weight w =
  function
      [] -> 0
    | t::ts -> 
	size_of_term_with_weight w t + size_of_terms_with_weight w ts;;
	  
(**
   This decides whether an uncoditional rule is size-decreasing.
*)
let is_size_decreasing w rl =
  is_nonduplicating rl 
    &&
  begin
    match rl with
      Rew (l,r,[]) -> 
	size_of_term_with_weight w l > size_of_term_with_weight w r
    | _ -> false
  end
and is_size_decreasing_eq w rl =
  is_nonduplicating rl
    &&
  begin
    match rl with
      Rew (l,r,[]) -> 
	size_of_term_with_weight w l >= size_of_term_with_weight w r
    | _ -> false
  end;;
    
(**
   This collects unconditional [->] rules from a given CTRS.
*)
let unconditional spec =
  let rec get_uncond =
    function
	[] -> []
      |	(Rew(l,r,[]) as rl)::rs -> rl::(get_uncond rs)
      |	_::rs -> get_uncond rs
  in
  mkFromRules (get_uncond spec.rules);;

(**
   This decomposes a condition c(s_1,...,s_n) = c(t_1,...,t_n)
   to s_1 = t_1,...,s_n = t_n, whre c is a constructor,
   and removes s = x with s a ground strongly irreducible term
   by replacing x with s.
   This works Arrow conditions only.
*)
let decompose_condition_of_rule spec ds rl =
  let rec decompose l r cs0 cs =
    begin
      match cs with
        [] -> (l,r,cs0)
    | (c1::cs1) ->
        begin
          match c1 with
           Arrow (Node (FSym f,ts),Node (FSym g,us)) ->
	      if f = g && not (List.mem f ds)
              then
                let cs2 = mk_conditions ts us in
                decompose l r cs0 (cs2@cs1)
              else
                decompose l r (cs0@[c1]) cs1
        | Arrow (s,Node (Var x,_)) ->
             if is_ground_strongly_irreducible_term spec s
             then
               let sigma = [(x,s)] in
               decompose (apply_subst_to_term sigma l) (apply_subst_to_term sigma r) 
                            (List.map (apply_subst_to_condition sigma) cs0) (List.map (apply_subst_to_condition sigma) cs1) 
             else
               decompose l r (cs0@[c1]) cs1
         | Arrow _ ->
             decompose l r (cs0@[c1]) cs1 
         | _ -> raise (Failure "unexpected")
       end
    end
  and mk_conditions ts us =
    begin
      match (ts,us) with
        ([],[]) -> []
     | (t1::ts1,u1::us1) -> (Arrow (t1,u1))::(mk_conditions ts1 us1)
     | _ -> raise (Failure "unexpected")
    end
  in
  match rl with
      Rew (l,r,c) ->
        let (l1,r1,c1) =  decompose l r [] c in
        Rew (l1,r1,c1)
    | RelRew (l,r,c) -> 
        let (l1,r1,c1) =  decompose l r [] c in
        RelRew (l1,r1,c1)
;;
(*
let decompose_condition_of_rule ds rl =
  let rec decompose c =
    begin
      match c with
	Arrow (Node (FSym f,ts),Node (FSym g,us)) ->
	  if f = g && not (List.mem f ds)
	  then
	    mk_conditions ts us
	  else
	    [c]
      |	Arrow (_,_) -> [c]
      |	_ -> raise (Failure "unexpected")
    end
  and mk_conditions ts us =
    begin
      match (ts,us) with
	([],[]) -> []
      |	(t::ts,u::us) -> (decompose (Arrow (t,u))) @ (mk_conditions ts us)
      |	_ -> raise (Failure "unexpected")
    end
  in
  begin
    match rl with
      Rew (l,r,c) -> Rew (l,r,List.concat (List.map decompose c))
    | RelRew (l,r,c) -> RelRew (l,r,List.concat (List.map decompose c))
  end;;
*)
      
(**
   This decomposes conditions of all rules.
*)
let decompose_conditions spec : ctrs =
  let ds = get_dsym spec.rules in
  let rs = List.map (decompose_condition_of_rule spec ds) spec.rules in
  { var = spec.var;
    theories = spec.theories ;
    rules = rs ;
    strategy = spec.strategy ; 
    others = spec.others ;
    ctype = spec.ctype ;
  } ;;

(**
   This rearranges [CONDITIONTYPE] if intpu is a TRS.
*)
let to_trs spec = 
   if is_unconditional spec 
   then
     { var = spec.var;
       theories = spec.theories;
       rules = spec.rules;
       strategy = spec.strategy;
       others = spec.others;
       ctype = None;
     } 
   else
     spec;;
