(**
   This module provides basic data structures for variables.
   @author Naoki Nishida
   @since 2010/12/17
   @since 2017/01/06
*)

open Util;;

(**
   [vid] is a data type for variables.
*)
type vid = string * int option;;
  
(**
   This converts variable names (in [vid]) to strings, 
   e.g., [string_of_vid ("x",None) = "x"],
   [string_of_vid ("x",Some 1) = "x1"],
   [string_of_vid ("x",Some -1) = "x-1"].
*)
let string_of_vid (s,n) =
  begin
    match n with
      None -> s
    | Some m -> s^(string_of_int m)
  end;;
    
(**
   This converts [vid] to [string] for debug.
*)
let string_of_vid0 (s,n) =
  "("^s^","
  ^
  begin
    match n with
      None -> "None"
    | Some m -> "Some "^(string_of_int m)
  end
  ^")";;

(**
   This converts strings to variable names (in [vid]),
   e.g.,  
   [vid_of_string "x" = ("x",None)],
   [vid_of_string "x0" = ("x",Some 0)],
   [vid_of_string "x00" = ("x0",Some 0)],
   [vid_of_string "x1" = ("x",Some 1)],
   [vid_of_string "x01" = ("x0",Some 1)],
   [vid_of_string "x-1" = ("x",Some -1)].
*)
let vid_of_string s = 
  let rec arrange_zero ss1 ss2 =
    begin
      match ss2 with
	[] -> (implode ss1,None)
      | [s31] ->
	  let ss4 = implode ss2 in
	  begin
	    try 
	      let n = int_of_string ss4 in
	      (implode ss1,Some n)
	    with
	      Failure "int_of_string" ->
		raise (Failure "Variable.vid_of_string: unexpected")
	  end
      | s31::s32::ss3 ->
	  if 
	    s31 = '0' 
	  then 
	    arrange_zero (ss1@[s31]) (s32::ss3)
	  else 
	    let ss4 = implode ss2 in
	    begin
	      try
		let n = int_of_string ss4 in
		(implode ss1,Some n)
	      with
		Failure "int_of_string" ->
		  raise (Failure "Variable.vid_of_string: unexpected")
	    end
    end
  and decompose ss1 ss2 =
    begin
      match ss1 with
	[] -> arrange_zero [] ss2
      | (s3::ss3) ->
	  if
	    '0' <= s3 && s3 <= '9'
	  then 
	    decompose ss3 (s3::ss2)
	  else 
	    arrange_zero (List.rev ss1) ss2
    end
  in
  let v = decompose (List.rev (explode s)) [] in
(*
  print_string ("(COMMENT vid_of_string "^s^" "^(string_of_vid0 v)^")\n");
*)
  v;;	

(**
   This converts [vid list] to [string].
*)
let rec string_of_vids =
  function
      [] -> ""
    | [x] -> string_of_vid x
    | x::xs -> (string_of_vid x)^" "^(string_of_vids xs);;
	  
(**
   This converts [vid list] to [string].
*)
let rec string_of_vids_with_comma =
  function
      [] -> ""
    | [x] -> string_of_vid x
    | x::xs -> (string_of_vid x)^", "^(string_of_vids_with_comma xs);;
	  
(**
   This produces [vid] data from pair of strings and integers.
*)
let vid (x,n) = (x,Some n);;
    
(**
   This returns string parts of [vid] data.
*)
let ofString x = fst x;;

(**
   This returns integer parts of [vid] data.
*)
let ofInt x = snd x;;
    
(**
   This is a variant of [ofInt].
*)
let ofIndex = ofInt;;

(**
   This looks for the maximum index in variable list.
*)
let rec get_max vname n =
  function
      [] -> n
    | (x,m)::xs -> 
	if x = vname 
	then
	  begin
	    match m with
	      None -> get_max vname n xs
	    | Some k -> if k > n then get_max vname k xs else get_max vname n xs
	  end
	else get_max vname n xs;; 
    
(**
   This looks for the maximum index in the siginature [fs] 
   that is a [string] list.
*)
let rec get_max_in_fids vname n =
  function
      [] -> n
    | f::fs -> 
	let (x,m) = f in
	if x = vname 
	then
	  begin
	    match m with
	      None -> get_max_in_fids vname n fs
	    | Some k -> 
		if k > n then get_max_in_fids vname k fs 
		else get_max_in_fids vname n fs
	  end
	else get_max_in_fids vname n fs;; 
    
(**
   This is an object to produce a fresh variable that are not in [xs].
*)
class fresh ((vname_init : string), xs_init, fs_init) =
  object
    val mutable vname = vname_init;
    val mutable fs = fs_init;
    val mutable xs = xs_init;
    val mutable max = 
      let n1 = get_max vname_init 0 xs_init in
      get_max_in_fids vname_init n1 fs_init;
    method reset (v',xs',fs') =
      begin
	vname <- v';
	xs <- xs';
	fs <- fs';
	max <- 
	  let n1 = get_max vname 0 xs in
	  get_max_in_fids vname n1 fs;
      end;
    method generate = 
      begin
	max <- max + 1;
	(vname,Some max)
      end;
    method add_vars vs =
      let m = get_max vname max vs in
      xs <- xs @ vs;
      max <- m;
    method add_var v =
      let m = get_max vname max  [v] in
      xs <- xs @ [v];
      max <- m;
    method current_vname = vname;
    method current_max = max;
  end;;

(**
   This is a global object to produce fresh variables.
*)
let gen_var = new fresh ("x",[],[]);;

(**
   This initializes [gen_var] with using [vname_init], [xs_init] and [fs_init].
*)
let init_gen_var (vname_init, xs_init, fs_init) =
  gen_var#reset (vname_init, xs_init, fs_init);;
